//---------------------------------------------------------------------------
// see mixmastdlg.cpp for notes on load-time vs run-time binding
// MixDir() is in global.cpp
#include <vcl\vcl.h>
#include <vcl\registry.hpp>
#pragma hdrstop

#include <assert.h>
#include "Mixmaster.h"
#include "console.h"  // for RunConsoleApp
#include "global.h" // for AppDir
#include "wiper.h"
#include "fileman.h"

HINSTANCE hMixLib=NULL;
HINSTANCE hSSLLib=NULL;
AnsiString mretbuf;
//---------------------------------------------------------------------------
AnsiString __fastcall MixPublicRing()
{
   AnsiString name=MixDir()+"pubring.mix";
   if (!FileExists(name))
   {
      MsgBox("Unable to locate mixmaster public keyring!","Mixmaster Error",MB_OK);
      return "";
   }
   return name;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixGetNewRemailerKeys(TStrings *KeyDoc)
{
 // returns description lines of new, not disabled, keys
   AnsiString desc,mtyp2txt=MixDir()+MTyp2Txt;
   TStringList *newkeys = new TStringList();
   TMixDb *curdb = new TMixDb(MixPublicRing());
   TMixDb *newdb = new TMixDb(KeyDoc);
   TStringList *type2 = new TStringList(); type2->Sorted = true;
   if (FileExists(mtyp2txt)) type2->LoadFromFile(mtyp2txt);
   TMixKey *key = newdb->FirstKey();
   while (key)
   {
      desc = key->Description;
      if (curdb->FindDescrip(desc)<0 && type2->IndexOf("#"+desc))
         newkeys->Add(desc);
      key = key->Next;
   }
   mretbuf = newkeys->Text;
   delete newdb;
   delete curdb;
   delete newkeys;
   delete type2;
   return mretbuf;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixAddNewRemailerKeys(TStrings *KeyDoc)
{
   TStringList *type2 = new TStringList();
   if (FileExists(MixDir()+MTyp2Txt)) type2->LoadFromFile(MixDir()+MTyp2Txt);
   TMixDb *curdb = new TMixDb(MixPublicRing());
   TMixDb *newdb = new TMixDb(KeyDoc);
   TMixKey *key = newdb->FirstKey();
   // find new keys and add to curdb
   while (key)
   {
      AnsiString desc = key->Description;
      if (curdb->FindDescrip(desc)<0 && type2->IndexOf("#"+desc)<0)
         curdb->AddKey(new TMixKey(key));
      key = key->Next;
   }
   mretbuf = curdb->SaveToString();
   delete newdb;
   delete curdb;
   delete type2;
   return mretbuf;
}
//---------------------------------------------------------------------------
bool __fastcall MixRingLineOk(AnsiString &line)
{
   int fld=0;
   bool ok=true;
   AnsiString data;
   while (ok && !(data=GetParam(fld,line," ")).IsEmpty())
   {
      if (fld==1 && data.Pos("@")<1) ok = false;
      if (fld==3 && data.Pos(".")<1) ok = false;
      fld++;
   }
   if (fld!=4 && fld!=5) ok = false;
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall MixListLineOk(AnsiString &line)
{
   return (line.Length()==44 && line[34]==':' && line[41]=='.' && line[44]=='%');
}
//---------------------------------------------------------------------------
bool __fastcall InType2(AnsiString &remailer, TStrings *list)
{
   bool ok = false;
   for (int i=0; i<list->Count; i++)
   {
      if (list->Strings[i]==remailer || list->Strings[i]=="#"+remailer)
      {
         ok = true;
         break;
      }
   }
   return ok;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixGenType2()
{
   int i;
   AnsiString line,mtyp2txt=MixDir()+MTyp2Txt,mringtxt=MixDir()+MRingTxt;

   TStringList *mring = new TStringList();
   if (FileExists(mringtxt)) mring->LoadFromFile(mringtxt);
   TStringList *type2 = new TStringList();
   if (FileExists(mtyp2txt)) type2->LoadFromFile(mtyp2txt);

   // remove all but disabled keys from type2.list
   i=0; while (i<type2->Count)
   {
      if (type2->Strings[i].SubString(1,1)!="#") type2->Delete(i);
      else i++;
   }
   // add keys from pubring.mix
   for (i=0; i<mring->Count; i++)
   {
      line = mring->Strings[i];
      if (!MixRingLineOk(line)) continue;
      if (!InType2(line,type2)) type2->Add(line);
   }

   mretbuf = type2->Text;
   delete type2;
   delete mring;
   return mretbuf;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixListKeyNotFound()
{
   AnsiString line,mlisttxt=MixDir()+MListTxt,mtyp2txt=MixDir()+MTyp2Txt;
   TStringList *mlist = new TStringList();
   TStringList *type2 = new TStringList();
   TStringList *nokeys = new TStringList();
   if (FileExists(mlisttxt)) mlist->LoadFromFile(mlisttxt);
   if (FileExists(mtyp2txt)) type2->LoadFromFile(mtyp2txt);
   // Get all short names in mlist.txt
   int i=0; while (i<mlist->Count)
   {
      line = mlist->Strings[i];
      if (MixListLineOk(line)) { mlist->Strings[i] = GetParam(0,line," "); i++; }
      else mlist->Delete(i);
   }
   // Get all short names in type2.list
   i=0; while (i<type2->Count)
   {
      line = type2->Strings[i];
      if (MixRingLineOk(line)) { type2->Strings[i] = GetParam(0,line," "); i++; }
      else type2->Delete(i);
   }
   // if a short name is in mlist and not type2, add it to nokeys
   for (i=0; i<mlist->Count; i++)
      if (type2->IndexOf(mlist->Strings[i])<0) nokeys->Add(mlist->Strings[i]);

   mretbuf = nokeys->Text;
   delete nokeys;
   delete type2;
   delete mlist;
   return mretbuf;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixType2Doubles()
{
  int i,c;
  AnsiString s,id,line,mtyp2txt=MixDir()+MTyp2Txt;
  TStringList *dbls = new TStringList();
  TStringList *type2 = new TStringList();
  if (FileExists(mtyp2txt)) type2->LoadFromFile(mtyp2txt);
  // get key ids in format             keyid=qty
  //                        acc09ab3d244db883a5c330c3843c8e3=2
  for (i=0; i<type2->Count; i++)
  {
     if (type2->Strings[i].Pos("#")==1) continue;

     id = GetParam(0,type2->Strings[i]," ");
     if (dbls->Values[id].IsEmpty()) dbls->Values[id]="0";

     c = dbls->Values[id].ToIntDef(0)+1;
     dbls->Values[id] = AnsiString(c);
  }
  // get list of keyids (only) with doubles
  i=0; while(i<dbls->Count)
  {
     s = dbls->Strings[i];
     if (s.Pos("=1")==0)
     {
        dbls->Strings[i] = s.SubString(1,s.Pos("=")-1);
        i++;
     }
     else dbls->Delete(i);
  }
  // exit if nothing to do
  if (!dbls->Count)
  {
     delete type2;
     delete dbls;
     return "";
  }
  // create actual display lines
  TStringList *dub = new TStringList();
  for (i=0; i<dbls->Count; i++)
  {
     for (c=0; c<type2->Count; c++)
     {
        if (type2->Strings[c].Pos(dbls->Strings[i])>0)
           dub->Add(type2->Strings[c]);
     }
     dub->Add("");
  }
  mretbuf = dub->Text;
  delete dub;
  delete type2;
  delete dbls;
  return mretbuf;
}
//---------------------------------------------------------------------------
bool __fastcall mixdb_getnext(TStringList *key, TStrings *keylist)
{
   key->Clear();
   AnsiString keyline;
   while (keylist->Count && keylist->Strings[0].IsEmpty()) keylist->Delete(0);
   if (!keylist->Count) return false;

   while (keylist->Count)
   {
      keyline = keylist->Strings[0];
      key->Add(keyline); keylist->Delete(0);
      if (keyline=="-----End Mix Key-----")
      {
         return true;
      }
   }
   return false;
}
//---------------------------------------------------------------------------
__fastcall TMixDb::TMixDb(AnsiString KeyRing)
{
   lastkey = NULL;
   firstkey = NULL;
   keyring = KeyRing;
   if (!keyring.IsEmpty())
   {
      TStringList *key = new TStringList();
      TStringList *keyfile = new TStringList();
      keyfile->LoadFromFile(keyring);

      while (mixdb_getnext(key,keyfile))
         AddKey(new TMixKey(keyring,key));

      delete keyfile;
      delete key;
   }
   modified = false;
}
//---------------------------------------------------------------------------
__fastcall TMixDb::TMixDb(TStrings *MixKeys)
{
   lastkey = NULL;
   firstkey = NULL;
   if (MixKeys==NULL) return;
   TStringList *key = new TStringList();
   TStringList *mkeys = new TStringList();
   mkeys->Text = MixKeys->Text;
   while (mixdb_getnext(key,mkeys))
     AddKey(new TMixKey(keyring,key));
   delete mkeys;
   delete key;
}
//---------------------------------------------------------------------------
void __fastcall TMixDb::Clear()
{
   TMixKey *tmp,*del=firstkey;
   while (del)
   {
      tmp = del->Next;
      delete del; del=NULL;
      if (tmp) del = tmp->Next;
   }
   modified = false;
   firstkey = NULL;
   lastkey = NULL;
   keyring = "";
}
//---------------------------------------------------------------------------
__fastcall TMixDb::~TMixDb()
{
//   if (modified && !keyring.IsEmpty())
//      SaveToFile(keyring);
   Clear();
}
//---------------------------------------------------------------------------
void __fastcall TMixDb::SaveToFile(AnsiString &FileName)
{
   TMixKey *key = firstkey;
   TFileStream *f = new TFileStream(FileName,fmCreate);
   while (key)
   {
      key->Write(f);
      key = key->Next;
   }
   modified = false;
   delete f;
}
//---------------------------------------------------------------------------
void __fastcall TMixDb::SaveToStrings(TStrings *strings)
{
   strings->Clear();
   TMixKey *key = firstkey;
   while (key)
   {
      strings->AddStrings(key->Key);
      if (!strings->Strings[strings->Count-1].IsEmpty())
         strings->Add("");
      key = key->Next;
   }
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMixDb::SaveToString()
{
   AnsiString keys;
   TStringList *strings = new TStringList();
   SaveToStrings(strings);
   keys = strings->Text;
   delete strings;
   return keys;
}
//---------------------------------------------------------------------------
int __fastcall TMixDb::AddKey(TMixKey *key)
{
   key->Next = NULL; key->Prev = NULL;
   if (!firstkey) { firstkey = key; lastkey = key; }
   else
   {
      key->Prev = lastkey;
      lastkey->Next = key;
      lastkey = key;
   }
   modified = true;
   return GetCount();
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::KeyWithIdText32(AnsiString Text32)
{
   TMixKey *key = firstkey;
   while (key && key->IdText32!=Text32)
      key = key->Next;
   return key;
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::FirstKey()
{
   return firstkey;
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::LastKey()
{
   return lastkey;
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::NextKey(TMixKey *FromThis)
{
   return FromThis==NULL ? NULL:FromThis->Next;
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::PrevKey(TMixKey *FromThis)
{
   return FromThis==NULL ? NULL:FromThis->Prev;
}
//---------------------------------------------------------------------------
TMixKey* TMixDb::GetKey(int pos)
{
   int p=0;
   TMixKey *key=firstkey;
   while (key && p<pos)
   {
      key = key->Next;
      p++;
   }
   return key;
}
//---------------------------------------------------------------------------
int __fastcall TMixDb::Delete(int pos)
{
   TMixKey *key=firstkey;//,*prev=NULL,*next=NULL;
   for (int i=0; i<pos; i++) if (key) key = key->Next;
   if (key)
   {
      if (key->Prev) key->Prev->Next = key->Next;
      if (key->Next) key->Next->Prev = key->Prev;
      modified = true;
      delete key;
   }
   return GetCount();
}
//---------------------------------------------------------------------------
void __fastcall TMixDb::OrderBy(TStrings *type2)
{
   int pos;
   AnsiString descrip;
   TMixKey *key,*first=NULL,*last=NULL;
   for (int i=0; i<type2->Count; i++)
   {
      descrip = type2->Strings[i];
      pos = FindDescrip(descrip);
      if (pos<0) pos = FindDescrip("#"+descrip);
      if (pos>=0)
      {
         key = GetKey(pos); if (!key) continue;
         key = new TMixKey(key); // make copies
         if (!first) { first = key; last = key; }
         else
         {
            key->Prev = last;
            last->Next = key;
            last = key;
         }
      }
   }
   Clear(); // delete originals
   firstkey = first;
   lastkey = last;
   modified = true;
}
//---------------------------------------------------------------------------
int __fastcall TMixDb::Find(TMixKey* Key)
{
   int pos=-1,keypos=0;
   TMixKey *key=firstkey;
   while (key)
   {
      if (key->IdText32==Key->IdText32)
      {
         pos = keypos;
         break;
      }
      key = key->Next;
      keypos++;
   }
   return pos;
}
//sm(key->IdText32+"\n"+Key->IdText32);
//---------------------------------------------------------------------------
int __fastcall TMixDb::FindDescrip(AnsiString &Descrip)
{
   int pos=-1,keypos=0;
   TMixKey *key=firstkey;
   while (key)
   {
      if (key->Description==Descrip)
      {
         pos = keypos;
         break;
      }
      key = key->Next;
      keypos++;
   }
   return pos;
}
//---------------------------------------------------------------------------
int TMixDb::GetCount()
{
   int count=0;
   TMixKey *key=firstkey;
   while (key)
   {
      key = key->Next;
      count++;
   }
   return count;
}
//---------------------------------------------------------------------------
__fastcall TMixKey::TMixKey(AnsiString Keyring, TStringList *KeyLines)
{
   key = new TStringList();
   key->Text = KeyLines->Text;
   keyring = Keyring;
   Next = NULL;
   Prev = NULL;
}
//---------------------------------------------------------------------------
__fastcall TMixKey::TMixKey(TMixKey *Key)
{
   key = new TStringList();
   key->Text = Key->key->Text;
   keyring = Key->Keyring;
   Next = NULL;
   Prev = NULL;
}
//---------------------------------------------------------------------------
__fastcall TMixKey::~TMixKey()
{
   delete key;
}
//---------------------------------------------------------------------------
long __fastcall TMixKey::Write(TFileStream *Stream)
{
   AnsiString keytxt = key->Text + "\r\n";
   return Stream->Write(keytxt.c_str(),keytxt.Length());
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetDescription()
{
   return key->Count>0 ? key->Strings[0]:AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetIdText32()
{
   return key->Count ? GetParam(2,key->Strings[0]," "):AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetName()
{
   return key->Count ? GetParam(0,key->Strings[0]," "):AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetUserId()
{
   return key->Count ? GetParam(1,key->Strings[0]," "):AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetVersion()
{
   return key->Count ? GetParam(3,key->Strings[0]," "):AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetCaps()
{
   return key->Count ? GetParam(4,key->Strings[0]," "):AnsiString("");
}
//---------------------------------------------------------------------------
AnsiString TMixKey::GetKeyring()
{
   return keyring;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
bool _fastcall HasMixmaster()
{
   return hMixLib>0 && hSSLLib>0;
}
//---------------------------------------------------------------------------
bool __fastcall HasMix()
{
   return HasMixmaster();
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixDir()
{
   AnsiString mixdir = "";
   TRegistry *reg = new TRegistry();
   if (reg->OpenKey(MixKey,false)) mixdir = reg->ReadString("MixDir");
   delete reg;
   if ( mixdir.IsEmpty()) mixdir = getenv("MIXDIR");
   if (!mixdir.IsEmpty() && mixdir[mixdir.Length()]!='\\') mixdir += "\\";
   return mixdir.LowerCase();
}
//---------------------------------------------------------------------------
int buf_readfile(BUFFER *buffer, char *filename)
{
   int err = -1;
   buf_reset(buffer);
   TFileStream *f;
   if (!FileExists(filename))
   {
      MsgBoxHwnd(NULL,
         s("The file "+s(filename)+" does not exist!\n")+
         s("Unable to complete operation."),
        "QS Pgp Error",
         MB_OK
      );
      return err;
   }
   try
   {
      f = new TFileStream(filename,fmOpenRead);
   }
   catch (...)
   {
      MsgBoxHwnd(NULL,
         s("I was unable to open "+s(filename)+" !\n")+
         s("Please make sure it is not in use by another program, and try\n")+
         s("the operation again."),
        "QS Pgp Error",
         MB_OK
      );
      f = NULL;
   }
   if (f)
   {
      char buf[BUFSIZE] = { 0 };
      while (true)
      {
         if (!f->Size) { err = 0; break; }

         int n = f->Read(buf,BUFSIZE); if (n > 0) err = 0;
         buf_append(buffer, (byte*)buf, n);
         if (n < BUFSIZE) break;

         if (buffer->length > BUFFER_MAX)
         {
            MsgBox("Message file too large. Giving up.","Error",MB_OK);
            return(1);
         }
      }
      delete f;
   }
   return err;
}
//---------------------------------------------------------------------------
int buf_writefile(BUFFER *buffer, char *filename, char* mode) // mode = "wb" or "wt"
{
   int err = -2;
   FILE *f = fopen(filename,mode);
   if (f) { err = -1;
      int n = (int)fwrite(buffer->data, 1, buffer->length, f)==buffer->length ? 0:-1;
      if (n>0) err=0;
      fclose(f);
   }
   return err;
/* int cnt=0;
   TFileStream *f = new TFileStream(MixTmp,fmCreate);
   if (f)
   {
      cnt = f->Write(buffer->data,buffer->length);
      delete f;
   }
   return cnt;
*/
}
//---------------------------------------------------------------------------
/*
BUFFER* __fastcall text2buf(AnsiString text)
{
   BUFFER *buf=buf_new();
   for (int i=0; i<text->Count; i++)
   {
      buf_appends(buf,text->Strings[i].c_str());
      buf_nl(buf);
   }
   return buf;
}
*/
//---------------------------------------------------------------------------
BUFFER* __fastcall buf_new(void)
{
   if (!hMixLib) return NULL;
   BUFFER* ret = NULL; BUF_NEW bufnew;
   if ((bufnew = (BUF_NEW)GetProcAddress(hMixLib,"buf_new"))!=NULL)
      ret = (bufnew)();
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_free(BUFFER *buf)
{
   if (!hMixLib) return NULL;
   int ret = NULL; BUF_FREE buffree;
   if ((buffree = (BUF_FREE)GetProcAddress(hMixLib,"buf_free"))!=NULL)
      ret = (buffree)(buf);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_reset(BUFFER *buf)
{
   if (!hMixLib) return NULL;
   int ret = NULL; BUF_RESET bufreset;
   if ((bufreset = (BUF_RESET)GetProcAddress(hMixLib,"buf_reset"))!=NULL)
      ret = (bufreset)(buf);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_append(BUFFER *buffer, byte *mess, int len)
{
   if (!hMixLib) return NULL;
   int ret = NULL; BUF_APPEND bufappend;
   if ((bufappend = (BUF_APPEND)GetProcAddress(hMixLib,"buf_append"))!=NULL)
      ret = (bufappend)(buffer,mess,len);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_sets(BUFFER *buffer, char *s)
{
   buf_clear(buffer);
   return (buf_appends(buffer, s));
}
//---------------------------------------------------------------------------
void __fastcall buf_move(BUFFER *dest, BUFFER *src)
{
   if (!hMixLib) return;
   BUF_MOVE bufmove;
   if ((bufmove = (BUF_MOVE)GetProcAddress(hMixLib,"buf_move"))!=NULL)
       (bufmove)(dest,src);
}
//---------------------------------------------------------------------------
int __fastcall mix_encrypt(int type, BUFFER *message, char *chain, int numcopies,
   BUFFER *feedback)
{
   if (!hMixLib) return NULL;
   int ret=NULL; MIX_ENCRYPT mixencrypt;
   if ((mixencrypt = (MIX_ENCRYPT)GetProcAddress(hMixLib,"mix_encrypt"))!=NULL)
      ret = (mixencrypt)(type,message,chain,numcopies,feedback);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall mix_encrypt_exe(int type, BUFFER *message, char *chain, int numcopies,
   BUFFER *feedback, bool *cancel, TProgressBar *progressbar)
{
   int ret=NULL; if (!hMixLib) return ret;

   AnsiString TheChain,TheCopies;
   TStringList *output = new TStringList();

   TheChain = chain!=NULL && strlen(chain)>0 ? "-l \""+AnsiString(chain)+"\" ":AnsiString("");
   TheCopies = numcopies > 1 ? "-c "+AnsiString(numcopies)+" " : AnsiString("");

   buf_writefile(message,MixTmp.c_str(),"wt");
   ret = RunConsoleApp(MixDir()+MixExe+" "+TheCopies+TheChain+MixTmp,output,SW_HIDE,
      cancel, progressbar);
   buf_sets(feedback,output->Text.c_str());
   simplewipe(MixTmp,10);
   delete output;
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall mix_send()
{
   if (!hMixLib) return NULL;
   int ret = NULL; MIX_SEND mixsend;
   if ((mixsend = (MIX_SEND)GetProcAddress(hMixLib,"mix_send"))!=NULL)
      ret = (mixsend)();
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall pgp_mailenc(int mode, BUFFER *message, char *sigid,
		BUFFER *pass, char *pubring, char *secring)
{
   if (!hMixLib) return NULL;
   int ret = NULL; PGP_MAILENC_BUF pgpmailenc;
   if ((pgpmailenc = (PGP_MAILENC_BUF)GetProcAddress(hMixLib,"pgp_mailenc"))!=NULL)
      ret = (pgpmailenc)(mode,message,sigid,pass,pubring,secring);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall pgp_encrypt(int mode, BUFFER *message, BUFFER *encr, BUFFER *sigid,
		BUFFER *pass, char *pubring, char *secring)
{
   if (!hMixLib) return NULL;
   int ret = NULL; PGP_ENCRYPT_BUF pgpencrypt;
   if ((pgpencrypt = (PGP_ENCRYPT_BUF)GetProcAddress(hMixLib,"pgp_encrypt"))!=NULL)
      ret = (pgpencrypt)(mode,message,encr,sigid,pass,pubring,secring);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall pgp_decrypt(BUFFER *message, BUFFER *pass, BUFFER *sig, char *pubring,
		char *secring)
{
   if (!hMixLib) return NULL;
   int ret = NULL; PGP_DECRYPT pgpdecrypt;
   if ((pgpdecrypt = (PGP_DECRYPT)GetProcAddress(hMixLib,"pgp_decrypt"))!=NULL)
      ret = (pgpdecrypt)(message,pass,sig,pubring,secring);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall rnd_add(byte *b, long len)
{
   if (!hMixLib) return NULL;
   int ret = NULL; RND_ADD rndadd;
   if ((rndadd = (RND_ADD)GetProcAddress(hMixLib,"rnd_add"))!=NULL)
      ret = (rndadd)(b,len);
   return ret;
}
//---------------------------------------------------------------------------
void __fastcall rnd_time()
{
   if (!hMixLib) return;
   RND_TIME rndtime;
   if ((rndtime = (RND_TIME)GetProcAddress(hMixLib,"rnd_time"))!=NULL)
      (rndtime)();
}
//---------------------------------------------------------------------------
int __fastcall rnd_mouse(UINT i, WPARAM w, LPARAM l)
{
   if (!hMixLib) return -1;
   int ret = NULL; RND_MOUSE rndmouse;
   if ((rndmouse = (RND_MOUSE)GetProcAddress(hMixLib,"rnd_mouse"))!=NULL)
      ret = (rndmouse)(i,w,l);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_read(BUFFER *outmsg, FILE *infile)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_READ bufread;
   if ((bufread = (BUF_READ)GetProcAddress(hMixLib,"buf_read"))!=NULL)
      ret = (bufread)(outmsg,infile);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_write(BUFFER *buffer, FILE *out)
{
   if (!hMixLib) return NULL;
   int ret = NULL; BUF_WRITE bufwrite;
   if ((bufwrite = (BUF_WRITE)GetProcAddress(hMixLib,"buf_write"))!=NULL)
      ret = (bufwrite)(buffer,out);
   return ret;
}
//---------------------------------------------------------------------------
/*int __fastcall buf_write(BUFFER *buffer, FILE *out)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_WRITE bufwrite;
   if ((bufwrite = (BUF_WRITE)GetProcAddress(hMixLib,"buf_write"))!=NULL)
      ret = (bufwrite)(buffer,out);
   return ret;
}*/
//---------------------------------------------------------------------------
int __fastcall buf_getline(BUFFER *buffer, BUFFER *line)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_GETLINE bufgetline;
   if ((bufgetline = (BUF_GETLINE)GetProcAddress(hMixLib,"buf_getline"))!=NULL)
      ret = (bufgetline)(buffer,line);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_getheader(BUFFER *buffer, BUFFER *field, BUFFER *content)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_GETHEADER bufgetheader;
   if ((bufgetheader = (BUF_GETHEADER)GetProcAddress(hMixLib,"buf_getheader"))!=NULL)
      ret = (bufgetheader)(buffer,field,content);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_prepare(BUFFER *buffer, int size)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_PREPARE bufprepare;
   if ((bufprepare = (BUF_PREPARE)GetProcAddress(hMixLib,"buf_prepare"))!=NULL)
      ret = (bufprepare)(buffer,size);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_appendc(BUFFER *to, byte b)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_APPENDC bufappendc;
   if ((bufappendc = (BUF_APPENDC)GetProcAddress(hMixLib,"buf_appendc"))!=NULL)
      ret = (bufappendc)(to,b);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_appendi(BUFFER *to, int i)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_APPENDI bufappendi;
   if ((bufappendi = (BUF_APPENDI)GetProcAddress(hMixLib,"buf_appendi"))!=NULL)
      ret = (bufappendi)(to,i);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_appendl(BUFFER *to, long l)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_APPENDL bufappendl;
   if ((bufappendl = (BUF_APPENDL)GetProcAddress(hMixLib,"buf_appendl"))!=NULL)
      ret = (bufappendl)(to,l);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_cat(BUFFER *to, BUFFER *from)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_CAT bufcat;
   if ((bufcat = (BUF_CAT)GetProcAddress(hMixLib,"buf_cat"))!=NULL)
      ret = (bufcat)(to,from);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_clear(BUFFER *buffer)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_CLEAR bufclear;
   if ((bufclear = (BUF_CLEAR)GetProcAddress(hMixLib,"buf_clear"))!=NULL)
      ret = (bufclear)(buffer);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_appends(BUFFER *buffer, char *s)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_APPENDS bufappends;
   if ((bufappends = (BUF_APPENDS)GetProcAddress(hMixLib,"buf_appends"))!=NULL)
      ret = (bufappends)(buffer,s);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_get(BUFFER *buffer, BUFFER *to, int n)
{
  buf_clear(to);
  if (buffer->length - buffer->ptr < n)
    return (-1);
  buf_append(to, buffer->data + buffer->ptr, n);
  buffer->ptr += n;
  return (0);
/*
   if (!hMixLib) return -1;
   int ret = NULL; BUF_GET bufget;
   if ((bufget = (BUF_GET)GetProcAddress(hMixLib,"buf_get"))!=NULL)
      ret = (bufget)(buffer,to,n);
   return ret;
*/
}
//---------------------------------------------------------------------------
int __fastcall buf_rest(BUFFER *to, BUFFER *from)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_REST bufrest;
   if ((bufrest = (BUF_REST)GetProcAddress(hMixLib,"buf_rest"))!=NULL)
      ret = (bufrest)(to,from);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_nl(BUFFER *b)
{
   if (!hMixLib) return -1;
   return (buf_append(b,(byte*)"\n", 1));
//   int ret = NULL; BUF_NL bufnl;
//   if ((bufnl = (BUF_NL)GetProcAddress(hMixLib,"buf_nl"))!=NULL)
//      ret = (bufnl)(b);
//   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_set(BUFFER *to, BUFFER *from)
{
   if (!hMixLib) return -1;
   int ret = NULL; BUF_SET bufset;
   if ((bufset = (BUF_SET)GetProcAddress(hMixLib,"buf_set"))!=NULL)
      ret = (bufset)(to,from);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall decode(BUFFER *in, BUFFER *out)
{
   if (!hMixLib) return -1;
   int ret = NULL; DECODE decode;
   if ((decode = (DECODE)GetProcAddress(hMixLib,"decode"))!=NULL)
      ret = (decode)(in,out);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall encode(BUFFER *in, int linelen)
{
   if (!hMixLib) return -1;
   int ret = NULL; ENCODE encode;
   if ((encode = (ENCODE)GetProcAddress(hMixLib,"encode"))!=NULL)
      ret = (encode)(in,linelen);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall strifind(const char *string, const char *keyword)
{
   if (!hMixLib) return -1;
   int ret = NULL; STRIFIND strfindi;
   if ((strfindi = (STRIFIND)GetProcAddress(hMixLib,"strifind"))!=NULL)
      ret = (strfindi)(string,keyword);
   return ret;
}
//---------------------------------------------------------------------------
int __fastcall buf_getc(BUFFER *buffer)
{
  if (buffer->ptr == buffer->length) return (-1);
  else return (buffer->data[buffer->ptr++]);
}

int __fastcall buf_geti(BUFFER *b)
{
  int i;
  if (b->ptr + 2 > b->length) return (-1);
  i  = buf_getc(b) << 8;
  i += buf_getc(b);
  return (i);
}

long __fastcall buf_getl(BUFFER *b)
{
  long l;
  if (b->ptr + 4 > b->length) return (-1);
  l  = buf_getc(b) << 24;
  l += buf_getc(b) << 16;
  l += buf_getc(b) << 8;
  l += buf_getc(b);
  return (l);
}
//---------------------------------------------------------------------------
int __fastcall buf_rewind(BUFFER *buffer)
{
  buffer->ptr = 0;
  return (0);
}
//---------------------------------------------------------------------------
int __fastcall buf_eq(BUFFER *b1, BUFFER *b2)
{
//  sanity_check(b1);
//  sanity_check(b2);
  if (b1->length != b2->length) return (0);
  if (b1->length > 0 && memcmp(b1->data, b2->data, b1->length) != 0) return (0);
  return (1);
}
//---------------------------------------------------------------------------
int mixfile(char *path, const char *name)
{
  assert(path != NULL && name != NULL);

//if (name[0] == '/' || MIXDIR == NULL)
  if (name[0] == '/' || MixDir().IsEmpty() == NULL)
    strncpy(path, name, PATHMAX);
  else {
//  strncpy(path, MIXDIR, PATHMAX);
    strncpy(path, MixDir().c_str(), PATHMAX);
    strncat(path, name, PATHMAX);
  }
  return (0);
}
//---------------------------------------------------------------------------
/*int PgpPacketType(BUFFER *b, long *len)
{
  int ctb;

  ctb = buf_getc(b);
  switch (ctb >> 6)
  {
     case 2:
       // old packet type
        switch (ctb & 3)
        {
           case 0:
             *len = buf_getc(b);
              break;
           case 1:
             *len = buf_geti(b);
              break;
           case 2:
             *len = buf_getl(b);
              break;
           case 3:
             *len = b->length - b->ptr;
              break;
        }
        return (ctb >> 2) & 15;
     case 3:
       // new packet type
       *len = buf_getc(b);
        if (*len >= 192 && *len <= 223)
            *len = (*len - 192) * 256 + buf_getc(b) + 192; else
        if (*len == 255)
            *len = buf_getl(b); else
        if (*len > 223)
            return (-2); // partial body length
        return (ctb & 63);
  }
  return (-1);
}*/
//---------------------------------------------------------------------------
/*int PgpIsConventional(BUFFER *b)
{
  long l;
  int type;
  type = PgpPacketType(b, &l);
  buf_rewind(b);
  return (type == PGP_ENCRYPTED);
}*/
//---------------------------------------------------------------------------
FILE *mix_openfile(const char *name, const char *a)
{
  char path[PATHMAX];

  mixfile(path, name);
  return (fopen(path, a));
}
//---------------------------------------------------------------------------






