//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <vcl\registry.hpp>
#pragma hdrstop

#include "NymWiz.h"
#include "Version.h"
#include "HeadMan.h"
#include "MixmastDlg.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TNymWizard *NymWizard;
//---------------------------------------------------------------------------
bool __fastcall RunNymWizard()
{
   NymWizard = new TNymWizard(Application->MainForm);
   NymWizard->ShowModal();
   bool DidIt = NymWizard->ModalResult==mrOk;
   delete NymWizard;
   return DidIt;
}
//---------------------------------------------------------------------------
__fastcall TNymWizard::TNymWizard(TComponent* Owner)
   : TForm(Owner)
{
   CenterDialog(this,dynamic_cast<TForm*>(Owner));

   LoadNymOptions();
   LoadNymServers(NymServerBox->Items);
   GetAppData(RBlkMemo->Lines,DefaultReplyBlock);

   bool hasNyms = AppDataExists(UserNyms);
   ModifyBtn->Enabled = hasNyms;
   DeleteBtn->Enabled = hasNyms;

   Panel1->BorderStyle = bsNone;
   BackBtn->Enabled = false;
   Notebook->PageIndex = 0;

   UpdateBox->Checked = Application->MainForm->Perform(WM_UPDATECHECKED,0,NULL);
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (ModalResult==mrCancel && MsgBoxHwnd(Handle,"Cancel config operation?",
      "Cancel",MB_YESNO)==IDNO)
   {
      CanClose = false;
      return;
   }
   if (options)
   for (int i=0; i<options->Count; i++)
      delete (TCheckBox*)options->Objects[i];
   delete options;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TNymWizard::GetNym()
{
   AnsiString nymloc,nymsrv,thenym;
   if (CreateBtn->Checked) { nymloc=NymLocal->Text; nymsrv=NymServerBox->Text; } else
   if (ModifyBtn->Checked) { ; } else
   if (DeleteBtn->Checked) { ; }
   if (thenym.IsEmpty() && nymloc.Length() && nymsrv.Length())
      thenym = nymloc+"@"+nymsrv;
   return thenym;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TNymWizard::GetNymName()
{
   AnsiString thename;
   if (CreateBtn->Checked) thename=FullName->Text; else
   if (ModifyBtn->Checked) thename=""; else
   if (DeleteBtn->Checked) thename="";
   return thename;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TNymWizard::GetNymOptions()
{
   AnsiString opts;
   for (int i=0; i<options->Count; i++)
      opts += options->Strings[i]+" ";
   return opts.TrimRight();
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::LoadNymOptions()
{
   int i;
   AnsiString nymops,op;
   options = new TStringList();
   nymops = DEFNYMOPS;
   TRegistry *reg = new TRegistry();
   if (reg->OpenKey(QuickKey,false))
   {
      nymops = reg->ReadString("DefNymOps");
      i=0; while ((op=GetParam(i++,nymops," "))!="")
         options->Add(op);
   }
   delete reg;
   // Initialize checkboxes
   AnsiString cap;
   int xpos=0,ypos=20;
   for (i=0; i<options->Count; i++)
   {
      cap = options->Strings[i];
      TCheckBox* ckb = new TCheckBox(this); ckb->Parent = Panel1;
      ckb->Checked = cap.SubString(1,1)=="+";
      ckb->Caption = cap.SubString(2,255);
      ckb->OnClick = CheckBoxClick;
      ckb->Left = xpos;
      ckb->Top = ypos;
      xpos += 80;
      if (xpos>340)
      {
         ypos += 30;
         xpos = 0;
      }
      options->Objects[i] = ckb;
      ckb->Show();
   }
   Notebook->PageIndex = 0;
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::CheckBoxClick(TObject *Sender)
{
   TCheckBox *chb = dynamic_cast<TCheckBox*>(Sender);
   for (int i=0; i<options->Count; i++)
   {
      if (options->Objects[i] == chb)
      {
         AnsiString s = options->Strings[i];
         if (s.Length()) s[1] = chb->Checked ? '+':'-';
         options->Strings[i] = s;
         break;
      }
   }
}
//---------------------------------------------------------------------------
bool __fastcall TNymWizard::LoadNymServers(TStrings *toList)
{
   toList->Clear();
   AnsiString rem,rlistfile = MixDir()+CListTxt;
   if (FileExists(rlistfile))
   {
      TStringList *txt = new TStringList();
      txt->LoadFromFile(rlistfile);
      for (int i=0; i<txt->Count; i++)
      {
          rem = txt->Strings[i]+" ";
          if (rem.Pos(" newnym "))
             toList->Add(HeadMan->GetDomain(HeadMan->GetAddress(rem)));
      }
      delete txt;
   }
   return toList->Count;
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::CNymChange(TObject *Sender)
{
   NextBtn->Enabled = NymLocal->Text.Length() && NymServerBox->Text.Length();
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::NextBtnClick(TObject *Sender)
{
   AnsiString s;
   switch (Notebook->PageIndex)
   {
      case StartPg:
         if (CreateBtn->Checked) Notebook->PageIndex = Create1; else
         if (ModifyBtn->Checked) Notebook->PageIndex = Modify1; else
         if (DeleteBtn->Checked) Notebook->PageIndex = Delete1;
         break;
      case Create3: // end of create
      case Modify3: // end of modify
      case Delete2: // end of delete
         Notebook->PageIndex = ReviewPg;
         break;
      case ReviewPg:
         Notebook->PageIndex = FinishPg;
         break;
      case FinishPg:
         if (GenerateConfigMessage())
         {
            ModalResult = mrOk;
         }
         break;
      default:
         Notebook->PageIndex++;
   }
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::BackBtnClick(TObject *Sender)
{
   switch (Notebook->PageIndex)
   {
      case 1:
      case 4:
      case 7:
         Notebook->PageIndex = StartPg;
         break;
      case ReviewPg:
         if (CreateBtn->Checked) Notebook->PageIndex = 3; else
         if (ModifyBtn->Checked) Notebook->PageIndex = 6; else
         if (DeleteBtn->Checked) Notebook->PageIndex = 8;
         break;
      default: Notebook->PageIndex--;
   }
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::NotebookPageChanged(TObject *Sender)
{
   BackBtn->Enabled = Notebook->PageIndex > 0;
   switch (Notebook->PageIndex)
   {
      case StartPg:
         Caption = NYMWIZARD;
         NextBtn->Enabled = true;
         ActiveControl = CreateBtn->Checked ? CreateBtn : ModifyBtn->Checked ? ModifyBtn : DeleteBtn;
         break;
      // Create Nym
      case 1:
         Caption = NYMWIZARD+" {create}";
         if (NymServerBox->Text.IsEmpty())
            NymServerBox->ItemIndex = NymServerBox->Items->IndexOf(DEFNYMSERVER);
         NextBtn->Enabled = !GetNym().IsEmpty();
         ActiveControl = NymLocal;
         break;
      case 2:
         ActiveControl = options->Count?(TWinControl*)options->Objects[0]:(TWinControl*)NULL;
         break;
      case 3:
         NextBtn->Caption = "Next";
         ActiveControl = RBlkMemo;
         RBlkMemo->SelStart = 0;
         break;
      // Modify Nym
      case 4:
         Caption = NYMWIZARD+" {modify}";
         break;
      case 5:
         break;
      case 6:
         break;
      // Delete Nym
      case 7:
         Caption = NYMWIZARD+" {delete}";
         break;
      case 8:
         break;
      // review and finish
      case ReviewPg:
         NextBtn->Caption = "Next";
         ActiveControl = Review;
         LoadReview();
         break;
      case FinishPg:
         NextBtn->Caption = "Finish";
         ActiveControl = Chain;
         break;
   }
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::PopSaveasDefaultItemClick(TObject *Sender)
{
   PutAppData(DefaultReplyBlock,RBlkMemo->Lines);
   MsgBoxHwnd(Handle,"Reply Block saved as default","Text Saved",MB_OK);
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::LoadReview()
{
   AnsiString s,last="";
   Review->Clear();
   Review->Lines->Add("NYM: "+GetNym());
   //Review->Lines->Add("Name: "+GetNymName());
   Review->Lines->Add("Options: "+GetNymOptions());
   Review->Lines->Add("Name: "+GetNymName());
   Review->Lines->Add("Reply-Blocks:");
   Review->Lines->Add("");
   for (int i=0; i<RBlkMemo->Lines->Count; i++)
   {
      s = RBlkMemo->Lines->Strings[i];
      if (last.IsEmpty() && s.IsEmpty()) continue;
      Review->Lines->Add(s);
      last = s;
   }
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::ChainChange(TObject *Sender)
{
   NextBtn->Enabled = !Chain->Text.IsEmpty();
}
//---------------------------------------------------------------------------
void __fastcall TNymWizard::UpdateBoxClick(TObject *Sender)
{
   Application->MainForm->Perform(WM_CHECKUPDATE,UpdateBox->Checked,NULL);
}
//---------------------------------------------------------------------------
bool __fastcall TNymWizard::GenerateConfigMessage()
{
   return true;
}
//---------------------------------------------------------------------------
