//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "pgp.h"
#include "config.h"
#include "mixmaster.h"
#include "global.h"
#include "headman.h"
TForm *ActMainForm=NULL;
//---------------------------------------------------------------------------
__fastcall TPgpDb::TPgpDb(AnsiString PubRing, AnsiString SecRing)
{
   int i;
   error = 0;
   lastkey = NULL;
   firstkey = NULL;
   AnsiString ring;
   TStringList *rings = new TStringList();
   i=0; while (!(ring=GetParam(i++,PubRing,"\n","")).IsEmpty()) rings->Add(ring);
   i=0; while (!(ring=GetParam(i++,SecRing,"\n","")).IsEmpty()) rings->Add(ring);
   BUFFER *key=buf_new();
   for (int r=0; r<rings->Count; r++)
   {
      buf_clear(key);
      AnsiString keyring = rings->Strings[r];
      if (keyring.IsEmpty()) continue;

      pgpdb=pgpdb_open(keyring.c_str(),NULL,false);
      if (pgpdb==NULL)
      {
       //MsgBoxHwnd(NULL,"Unable to open PGP keyring "+keyring,"QS Pgp Error",MB_OK);
         error = 1;
         break;
      }
      while (pgpdb_getnext(pgpdb,key)>=0)
      {
         TPgpKey *pgpkey = new TPgpKey(keyring,key);
         TPgpKey *matchkey=KeyWithIdText8(pgpkey->IdText8);
         switch (pgpkey->Type)
         {
            case PGP_PUBKEY:
            case PGP_PUBSUBKEY:
               if (!matchkey) AddKey(pgpkey);
               break;
            case PGP_SECKEY:
            case PGP_SECSUBKEY:
               if (matchkey && !matchkey->SecKey) matchkey->SecKey = pgpkey;
               else
                  MsgBox(
                     AnsiString("Your pgp keyrings are out of sync. The secret \n")+
                     AnsiString("key, '"+pgpkey->IdText8+"', does not have a corresponding \n")+
                     AnsiString("public key!"),
                    "PGP key Error",
                     MB_OK
                  );
               break;
            default:
               MsgBox(
                  AnsiString("The pgp key '"+pgpkey->IdText8+"' is corrupted. It \n")+
                  AnsiString("identifies itseld as neither a public nor secret key!"),
                 "PGP key Error",
                  MB_OK
               );
         }
      }
   }
   buf_free(key);
   delete rings;
}
//---------------------------------------------------------------------------
__fastcall TPgpDb::~TPgpDb()
{
   TPgpKey *key,*del=firstkey;
   while (del)
   {
      key = del->Next;
      delete del;
      del = key;
   }
   if (pgpdb)
   {
      buf_free(pgpdb->db);
      free(pgpdb);
   }
}
//---------------------------------------------------------------------------
void __fastcall TPgpDb::SaveToFile(AnsiString &FileName)
{
   BUFFER *buf = buf_new();
   bool ARMOR=ExtractFileExt(FileName).LowerCase()==".asc";
   TPgpKey *key = firstkey;
   while (key)
   {
      buf_cat(buf,key->Key);
      key = key->Next;
   }
   if (ARMOR) pgp_armor(buf,2);
   buf_writefile(buf,FileName.c_str(),ARMOR?"wt":"wb");
   buf_free(buf);
}
//---------------------------------------------------------------------------
void __fastcall TPgpDb::SaveSecretToFile(AnsiString &FileName)
{
   BUFFER *buf = buf_new();
   bool ARMOR=ExtractFileExt(FileName).LowerCase()==".asc";
   TPgpKey *key = firstkey;
   while (key)
   {
      if (key->IsPair) buf_cat(buf,key->SecKey->Key);
      key = key->Next;
   }
   if (ARMOR) pgp_armor(buf,2);
   buf_writefile(buf,FileName.c_str(),ARMOR?"wt":"wb");
   buf_free(buf);
}
//---------------------------------------------------------------------------
int __fastcall TPgpDb::AddKey(TPgpKey *key)
{
   if (!firstkey) { firstkey = key; lastkey = key; }
   else
   {
      key->Prev = lastkey;
      lastkey->Next = key;
      key->Next = NULL;
      lastkey = key;
   }
   return GetCount();
}
//---------------------------------------------------------------------------
int __fastcall TPgpDb::AppendDb(BUFFER *keybuf)
{
   pgpdb_append(pgpdb,keybuf);
   AddKey(new TPgpKey(keybuf));
   return GetCount();
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::KeyWithIdText8(AnsiString Text8)
{
   TPgpKey *key = firstkey;
   while (key && key->IdText8!=Text8)
      key = key->Next;
   return key;
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::FirstKey()
{
   return firstkey;
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::LastKey()
{
   return lastkey;
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::NextKey(TPgpKey *FromThis)
{
   return FromThis==NULL ? NULL:FromThis->Next;
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::PrevKey(TPgpKey *FromThis)
{
   return FromThis==NULL ? NULL:FromThis->Prev;
}
//---------------------------------------------------------------------------
TPgpKey* TPgpDb::GetKey(int pos)
{
   int p=0;
   TPgpKey *key=firstkey;
   while (key && p<=pos)
   {
      key = key->Next;
      p++;
   }
   return key;
}
//---------------------------------------------------------------------------
int __fastcall TPgpDb::Delete(int pos)
{
   TPgpKey *key=firstkey,*prev=NULL,*next=NULL;
   for (int i=0; i<pos; i++) if (key) key = key->Next;
   if (key)
   {
      if (key->Prev) key->Prev->Next = next;
      if (key->Next) key->Next->Prev = prev;
   }
   return GetCount();
}
//---------------------------------------------------------------------------
int __fastcall TPgpDb::Find(TPgpKey* Key)
{
   int fndkey=-1,keypos=0;
   TPgpKey *key=firstkey;
   while (key)
   {
      if (Key->IdText8==key->IdText8)
      {
         fndkey = keypos;
         break;
      }
      key = key->Next;
      keypos++;
   }
   return fndkey;
}
//---------------------------------------------------------------------------
int TPgpDb::GetCount()
{
   int count=0;
   TPgpKey *key=firstkey;
   while (key)
   {
      key = key->Next;
      count++;
   }
   return count;
}
//---------------------------------------------------------------------------
__fastcall TPgpKey::TPgpKey(AnsiString &Keyring,BUFFER *Key)
{
   key = buf_new(); buf_cat(key,Key);
   keyring = Keyring;
   GetKeyUserIds();
   SecKey = NULL;
   Next = NULL;
   Prev = NULL;
}
//---------------------------------------------------------------------------
__fastcall TPgpKey::TPgpKey(BUFFER *Key)
{
   key = buf_new(); buf_cat(key,Key);
   keyring = "";
   GetKeyUserIds();
   SecKey = NULL;
   Next = NULL;
   Prev = NULL;
}
//---------------------------------------------------------------------------
__fastcall TPgpKey::~TPgpKey()
{
   delete UserIds;
   if (SecKey) delete SecKey;
   buf_free(key);
}
//---------------------------------------------------------------------------
long __fastcall TPgpKey::Write(TFileStream *Stream)
{
   return Stream->Write(key->data,key->length);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPgpKey::GetDescription()
{
   AnsiString descrip=UserIds->Count ? UserIds->Strings[0]:s("no_user@_id");
   TStringList *adrs = new TStringList();
   THeadMan *hm = new THeadMan();
   hm->GetRemAddresses(adrs,descrip);
   if (!adrs->Count) adrs->Add(descrip);
//sm(descrip);
   descrip = "";
   for (int i=0; i<adrs->Count; i++)
   {
      if (i==0) descrip += adrs->Strings[0];
      else descrip += "/"+adrs->Strings[i];
   }
   descrip += " 0x"+IdText8;
//sm(descrip);
   delete hm;
   delete adrs;
   return descrip;
}
//---------------------------------------------------------------------------
int TPgpKey::GetType()
{
   int type = -1;
   int CTB = key->data[0];
//   if ((CTB & 24)==24) type = PGP_PUBKEY; else
//   if ((CTB & 20)==20) type = PGP_SECKEY;
   if (CTB==153) type = PGP_PUBKEY; else
   if (CTB==149) type = PGP_SECKEY;
   return type;
}
//---------------------------------------------------------------------------
int TPgpKey::PutMPI(BUFFER *buf, BUFFER *mpi)
{
  int l,i;
  l = mpi->length * 8;
  for (i = 7; i >= 0; i--)
    if (((mpi->data[0] >> i) & 1) == 1) {
      l -= 7 - i;
      break;
    }
  buf_appendi(buf, l);
  buf_cat(buf, mpi);
  return (l);
}
//---------------------------------------------------------------------------
int TPgpKey::GetMPI(BUFFER *mpi)
{
  int l=buf_geti(key);
  if (l <= 0 || key->ptr + (l + 7) / 8 > key->length) return (-1);
  buf_get(key, mpi, (l + 7) / 8);
  return (l);
}
//---------------------------------------------------------------------------
AnsiString TPgpKey::GetIdText8()
{
 	 byte c;
   AnsiString text8;
   BUFFER *id = buf_new();
   KeyId(id);
	 for (int i=0; i<4; i++)
   {
 	  	 c = id->data[4+i];
		 text8 += "0123456789ABCDEF" [ c >> 4  ];
		 text8 += "0123456789ABCDEF" [ c & 0xf ];
	 }
   buf_free(id);
   return text8;
}
//---------------------------------------------------------------------------
int TPgpKey::KeyId(BUFFER *keyid)
{
  BUFFER *i,*k;
  int version,algo,j;
  long ptr;

  i = buf_new();
  k = buf_new();
  ptr = key->ptr; key->ptr = 3;
  switch (version=buf_getc(key))
  {
     case 2:
     case 3:
        buf_getl(key); // timestamp
        buf_geti(key); // validity
        buf_getc(key); // algo
        GetMPI(i);
        break;
     case 4:
        buf_appendc(k, (char)version);
        buf_appendl(k, buf_getl(key));
        algo = buf_getc(key);
        buf_appendc(k, (char)algo);
        if (pgp_nummpi(algo) == 0) buf_rest(k, key); /* works for public keys only */
        else
           for (j = 0; j < pgp_nummpi(algo); j++)
           {
	            GetMPI(i);
 	            PutMPI(k,i);
           }
        buf_clear(i);
        buf_appendc(i, 0x99);
        buf_appendi(i, k->length);
        buf_cat(i, k);
        digest_sha1(i, i);
        break;
  }
  buf_clear(keyid);
  buf_append(keyid, i->data + i->length - 8, 8);
  key->ptr = ptr;
  buf_free(i);
  buf_free(k);
  return(0);
}
//---------------------------------------------------------------------------
int TPgpKey::GetKeyUserIds()
{
   int type;
   buf_rewind(key);
   BUFFER *pk=buf_new();
   UserIds = new TStringList();
   while ((type=pgp_getpacket(key,pk))>0)
   {
      if (type==PGP_USERID) UserIds->Add((char*)pk->data);
   }
   buf_free(pk);
   return UserIds->Count;
}
//---------------------------------------------------------------------------
byte TPgpKey::GetVersion()
{
   key->ptr = 3;
   return (byte)buf_getc(key);
}
//---------------------------------------------------------------------------
long TPgpKey::GetTimeStamp()
{
   key->ptr = 4;
   return buf_getl(key); // timestamp
}
//---------------------------------------------------------------------------
int TPgpKey::GetValidity()
{
   key->ptr = 8;
   return buf_geti(key); // version
}
//---------------------------------------------------------------------------
byte TPgpKey::GetAlgorithm()
{
   key->ptr = 10;
   return (byte)buf_getc(key); // RSA=1
}
//---------------------------------------------------------------------------
int TPgpKey::GetKeyBits()
{
   key->ptr = 11;
   BUFFER *tmp=buf_new();
   long bits = buf_geti(key);
   delete tmp;
   return bits;
}
//---------------------------------------------------------------------------
AnsiString TPgpKey::GetKeyring()
{
   return keyring;
}
//---------------------------------------------------------------------------
bool TPgpKey::GetIsPair()
{
   return SecKey!=NULL;
}
//---------------------------------------------------------------------------
void __fastcall SHA1(byte *b, int n, byte *m)
{
   if (!hSSLLib) return;
   _SHA1 sha1;
   if ((sha1 = (_SHA1)GetProcAddress(hSSLLib,"SHA1"))!=NULL) (sha1)(b,n,m);
}
//---------------------------------------------------------------------------
int pgp_readkeyring(BUFFER *keys, char *filename)
{
  BUFFER *armored = buf_new();
  int err = buf_readfile(armored,filename);
  if (err<0) return err;
  if (pgp_ispacket(armored))
  {
    buf_move(keys, armored);
    err = 0;
  }
  else
  {
    err = pgp_dearmor(armored, keys);
    if (err==0) err = ARMORED;
  }
  buf_free(armored);
  return (err);
}
//---------------------------------------------------------------------------
KEYRING *pgpdb_new(char *keyring, int filetype, BUFFER *encryptkey)
{
  KEYRING *keydb;

  keydb = (KEYRING*)malloc(sizeof(KEYRING));

  if (keydb == NULL) return NULL;
  keydb->db = buf_new();
  keydb->modified = 0;
  keydb->lock = NULL;
  strncpy(keydb->filename, keyring, sizeof(keydb->filename));
  keydb->filetype = filetype;
  if (encryptkey==NULL) keydb->encryptkey = NULL;
  else
  {
    keydb->encryptkey = buf_new();
    buf_set(keydb->encryptkey, encryptkey);
  }

  return (keydb);
}
//---------------------------------------------------------------------------
KEYRING *pgpdb_open(char *keyring, BUFFER *encryptkey, int writer)
{
  KEYRING *keydb;
  keydb = pgpdb_new(keyring, -1, encryptkey);
  if (writer) keydb->lock = lockfile(keyring);
  keydb->filetype = pgp_readkeyring(keydb->db, keyring);
// fix for empty file
// sm(s(keyring)+"   keydb: "+s(keydb->filetype));
#if 1
  if (keydb->filetype == -1)
  {
    pgpdb_close(keydb);
    return (NULL);
  }
#endif
  if (encryptkey && pgp_isconventional(keydb->db) &&
      pgp_decrypt(keydb->db, encryptkey, NULL, NULL, NULL) < 0)
  {
//  user_delpass();
    return (NULL);
  }
//  keydb->modified = false;
  return (keydb);
}
//---------------------------------------------------------------------------
int pgpdb_getnext(KEYRING *keydb, BUFFER *key)
{
  long ptr;
  int type;
  BUFFER *p=buf_new();

  buf_clear(key);
  bool quitnow=false,havekey=false;
  if ((type=pgp_getpacket(keydb->db,p))>-1)
  {
     //int t=type;
     pgp_packet(p,type); buf_cat(key, p);
     while ((ptr=keydb->db->ptr,type=pgp_getpacket(keydb->db, p))>0)
     {
        havekey = true;
        switch (type)
        {
           case PGP_SECKEY:
           case PGP_PUBKEY:
              keydb->db->ptr = ptr;
              quitnow = true;
              break;
           case PGP_PUBSUBKEY:
           case PGP_SECSUBKEY:
	            break;
           case PGP_USERID:
              //sm(String(t)+"   "+String((char*)p->data));
              break;
        }
        if (quitnow) break;

        pgp_packet(p, type);
        buf_cat(key, p);
     }
  }
  buf_free(p);
  return (havekey ? 0 : -1);
}
//---------------------------------------------------------------------------
int pgpdb_append(KEYRING *keydb, BUFFER *p)
{
//sm((int)keydb->db->length);
//  assert(keydb->lock);
  buf_cat(keydb->db, p);
  keydb->modified = 1;
  return (0);
}
//---------------------------------------------------------------------------
int pgpdb_close(KEYRING *keydb)
{
  int err = 0;
  BUFFER *db = keydb->db;
  if (keydb->modified)
  {
    FILE *f;
    if (keydb->encryptkey)
       pgp_encrypt(PGP_CONVENTIONAL | PGP_NOARMOR, keydb->db,
		     keydb->encryptkey, NULL, NULL, NULL, NULL);
    if (keydb->filetype == ARMORED)
       pgp_armor(keydb->db, 2);
    if (keydb->filetype == -1 || (f = mix_openfile(keydb->filename,
       keydb->filetype == ARMORED ? "w" : "wb")) == NULL)
      err = -1;
    else {
      err = ((int)fwrite(db->data,1,db->length,f)==db->length ? 0 : -1);
      fclose(f);
    }
  }
  if (keydb->lock) unlockfile(keydb->lock);
  buf_free(keydb->db);
  free(keydb);
  return (err);
}
//---------------------------------------------------------------------------
int pgp_packet(BUFFER *in, int type)
{
  int ctb;
  BUFFER *out;

  out = buf_new();
  ctb = 128 + (type << 2);
  if (in->length < 256 && type != PGP_PUBKEY && type != PGP_SECKEY &&
      type != PGP_SIG && type != PGP_PUBSUBKEY && type != PGP_SECSUBKEY
#ifdef MIMIC
      && type != PGP_ENCRYPTED
#endif
  )
  {
    buf_appendc(out, (char)ctb);
    buf_appendc(out, (char)in->length);
  }
#ifndef MIMIC
  else if (in->length < 65536)
#else
  else if ((type == PGP_PUBKEY || type == PGP_SECKEY || type == PGP_SIG
	   || type == PGP_SESKEY || type == PGP_PUBSUBKEY ||
	   type == PGP_SECSUBKEY) && in->length < 65536)
#endif
  {
    buf_appendc(out, (char)(ctb | 1));
    buf_appendi(out, in->length);
  } else {
    buf_appendc(out, (char)(ctb | 2));
    buf_appendl(out, in->length);
  }
  buf_cat(out, in);
  buf_move(in, out);
  buf_free(out);
  return (0);
}
//---------------------------------------------------------------------------
int pgp_packettype(BUFFER *b, long *len)
{
  int ctb = buf_getc(b);
  switch (ctb >> 6)
  {
     case 2:
        /* old packet type */
        switch (ctb & 3)
        {
           case 0: *len = buf_getc(b); break;
           case 1: *len = buf_geti(b); break;
           case 2: *len = buf_getl(b); break;
           case 3: *len = b->length - b->ptr; break;
        }
        return (ctb >> 2) & 15;
     case 3:
        /* new packet type */
        *len = buf_getc(b);
        if (*len >= 192 && *len <= 223) *len = (*len - 192) * 256 + buf_getc(b) + 192; else
        if (*len == 255) *len = buf_getl(b); else
        if (*len > 223) return (-2); /* partial body length */
        return (ctb & 63);
  }
  return (-1);
}
//---------------------------------------------------------------------------
int pgp_getpacket(BUFFER *in, BUFFER *p)
/*  returns <0 = error, >0 = packet type */
{
  long len;
  int type = pgp_packettype(in, &len);
  if (type>0) buf_get(in, p, len);
  return (type);
}
//---------------------------------------------------------------------------
   long lolf(BYTE CTB) // length-of-length field
   {
      long result = 1;
      if ((CTB & 1) == 1) result = 2; else
      if ((CTB & 2) == 2) result = 4; else
      if ((CTB & 3) == 3) result = 0;
      return result;
   }
   long getPacketSize(TFileStream *f, BYTE CTB)
   {
      long l;
      BYTE tmp;
      BYTE *b = (BYTE*)&l;
      UINT *w = (UINT*)&l;
      long result = 0;

      f->Read(&l,lolf(CTB));
      switch (lolf(CTB)) {
         case 1: result = *b; break;
         case 2:
            tmp=b[0]; b[0]=b[1]; b[1]=tmp;
            result = *w;
            break;
         case 4: result = l; break;
      }
      return result;
   }
   BYTE getCTB(TFileStream *f)
   {
      BYTE ctb;
      f->Read(&ctb,1);           // 1-------
      return ((ctb & 128) == 128) ? ctb : BYTE(0);
   }
   bool isSigCert(BYTE CTB)
   {
      return ((CTB & 8) == 8);   // --0010--
   }
   bool isKeyCert(BYTE CTB)
   {
      return ((CTB & 24) == 24); // --0110--
   }
   bool isUserId(BYTE CTB)
   {
      return ((CTB & 52) == 52); // --1101--
   }
   bool isTrustByte(BYTE CTB)
   {
      return ((CTB & 48) == 48); // --1100--
   }

void __fastcall SetKeyringTrust(TFileName &filename)
{
//   if (APPCPUNKCERTMIN==0) return;

   TFileStream *f;
   BYTE CTB=0,PREV,trust;
   char userid[1024];
   long nxt;
   int len;

   if (!FileExists(filename)) {
      ShowMessage("Unable to locate file...   \n"+filename);
      return;
   }

   f = new TFileStream(filename,fmOpenReadWrite);
   while (f->Position < f->Size)
   {
      PREV = CTB;
      if ((CTB = getCTB(f)) != 0)
      {
         len = getPacketSize(f,CTB);
//   sm(String(CTB)+": "+String(len));
if (isUserId(CTB))
{
   f->Read(&userid,len);
   userid[len]='\0';
   f->Seek(-len,soFromCurrent);
//   sm(String(userid)+": "+String(len));
}
         nxt = f->Position+len;
         if (isTrustByte(CTB))
         {
            f->Read(&trust,1);

            if (isUserId(PREV) && trust==0)
               trust = 129;       // 10000001
//            else
//            if (isKeyCert(PREV))
//               trust = 192;     // 11000000
//            else
//            if (isSigCert(PREV))
//               trust = 194;     // 11000010

            f->Seek(-1,soFromCurrent);
            f->Write(&trust,1);
         }
         f->Seek(nxt,soFromBeginning);

      }
      else
      {
         ShowMessage("Keyring Read Error: NOT a Cipher Type Byte (CTB)");
         break;
      }
   }
   delete f;
}
//---------------------------------------------------------------------------
int mpi_get(BUFFER *b, BUFFER *mpi)
{
  buf_clear(mpi);
  int l=buf_geti(b);
  if (l <= 0 || b->ptr + (l + 7) / 8 > b->length) return (-1);
  buf_get(b, mpi, (l + 7) / 8);
  return (l);
}
//---------------------------------------------------------------------------
int mpi_put(BUFFER *buf, BUFFER *mpi)
{
  int l,i;
  l = mpi->length * 8;
  for (i = 7; i >= 0; i--)
    if (((mpi->data[0] >> i) & 1) == 1) {
      l -= 7 - i;
      break;
    }
  buf_appendi(buf, l);
  buf_cat(buf, mpi);
  return (l);
}
//---------------------------------------------------------------------------
int pgp_nummpi(int algo)
{
  switch (algo)
  {
     case PGP_ES_RSA: return (2);
     case PGP_S_DSA : return (4);
     case PGP_E_ELG : return (3);
  }
  return (0);
}
//---------------------------------------------------------------------------
int digestmem_sha1(byte *b, int n, BUFFER *md)
{
  byte m[SHA_DIGEST_LENGTH];

  SHA1(b, n, m);
  buf_reset(md);
  buf_append(md, m, SHA_DIGEST_LENGTH);
  return (0);
}
//---------------------------------------------------------------------------
int digest_sha1(BUFFER *b, BUFFER *md)
{
  return (digestmem_sha1(b->data, b->length, md));
}
//---------------------------------------------------------------------------
int lock(FILE *f)
{
#ifndef WIN32
  struct flock lockstruct;
  lockstruct.l_type = F_WRLCK;
  lockstruct.l_whence = 0;
  lockstruct.l_start = 0;
  lockstruct.l_len = 0;
  return (fcntl(fileno(f), F_SETLKW, &lockstruct));
#else
  return (0);
#endif
}
//---------------------------------------------------------------------------
LOCK *lockfile(char *filename)
{
  LOCK *l;
  char name[LINELEN];

  strcpy(name, "lck");
  if (strchr(filename, '/'))
    strncat(name, strrchr(filename, '/'), LINELEN);
  else
    strncat(name, filename, LINELEN);
  l = (LOCK*)malloc(sizeof(LOCK));

  l->name = (char*)malloc(PATHMAX);
  mixfile(l->name, name);
  l->f = mix_openfile(l->name, "w+");
  if (l->f)
    lock(l->f);
  return (l);
}
//---------------------------------------------------------------------------
int unlock(FILE *f)
{
#ifndef WIN32

  struct flock lockstruct;

  lockstruct.l_type = F_UNLCK;
  lockstruct.l_whence = 0;
  lockstruct.l_start = 0;
  lockstruct.l_len = 0;
  return (fcntl(fileno(f), F_SETLKW, &lockstruct));
#else
  return (0);
#endif
}
//---------------------------------------------------------------------------
int unlockfile(LOCK *l)
{
  if (l->f) {
    unlock(l->f);
    fclose(l->f);
  }
  unlink(l->name);
  free(l->name);
  free(l);
  return (0);
}
//---------------------------------------------------------------------------
#define POLY 0X1864CFB

int pgp_armor(BUFFER *in, int mode)
     /* mode = 1: remailer message, 0: normal message, mode = 2: key 3=nym key
        4=nym signature */
{
  BUFFER *out;
  unsigned long crc = 0xB704CE;
  long p;
  int i;

#if 0
  /* CRC algorithm from RFC 2440 */
  for (p = 0; p < in->length; p++) {
    crc ^= in->data[p] << 16;
    for (i = 0; i < 8; i++) {
      crc <<= 1;
      if (crc & 0x1000000)
	crc ^= POLY;
    }
  }
#else
  /* pre-computed CRC table -- much faster */
  unsigned long table[256];
  unsigned long t;
  int q = 0;

  table[0] = 0;
  for (i = 0; i < 128; i++) {
    t = table[i] << 1;
    if (t & 0x1000000) {
      table[q++] = t ^ POLY;
      table[q++] = t;
    } else {
      table[q++] = t;
      table[q++] = t ^ POLY;
    }
  }
  for (p = 0; p < in->length; p++)
    crc = (crc << 8) ^ table[(in->data[p] ^ (crc >> 16)) & 255];
  //crc = crc << 8 ^ table[(in->data[p] ^ crc >> 16) & 255];
#endif
  encode(in, 64);

  out = buf_new();
  if (mode == 2 || mode == 3)
    buf_sets(out, begin_pgpkey);
  else if (mode == 4)
    buf_sets(out, begin_pgpsig);
  else
    buf_sets(out, begin_pgpmsg);
  buf_nl(out);
#if 1
  buf_appends(out, "Version: N/A\n");
#else
#ifdef MIMIC
  if (mode == 1 || mode == 3 || mode == 4)
    buf_appends(out, "Version: 2.6.3i\n");
  else
#endif
  {
    buf_appends(out, "Version: Mixmaster ");
    buf_appends(out, VERSION);
    buf_appends(out, " (OpenPGP module)\n");
  }
#endif
  buf_nl(out);
  buf_cat(out, in);
  buf_reset(in);
  buf_appendc(in, (byte)((crc >> 16) & 255));
  buf_appendc(in, (byte)((crc >> 8) & 255));
  buf_appendc(in, (byte)( crc & 255));
  encode(in, 0);
  buf_appendc(out, '=');
  buf_cat(out, in);
  buf_nl(out);
  if (mode == 2 || mode == 3)
    buf_appends(out, end_pgpkey);
  else if (mode == 4)
    buf_appends(out, end_pgpsig);
  else
    buf_appends(out, end_pgpmsg);
  buf_nl(out);

  buf_move(in, out);
  buf_free(out);
  return (0);
}
//---------------------------------------------------------------------------
int pgp_isconventional(BUFFER *b)
{
  int type;
  long l;

  type = pgp_packettype(b, &l);
  buf_rewind(b);
  return (type == PGP_ENCRYPTED);
}
//---------------------------------------------------------------------------
int strleft(const char *string, const char *keyword)
{
  register unsigned int i;

  for (i=0; keyword[i]!='\0'; i++)
     if (string[i] != keyword[i]) return 0;
  return 1;
}
//---------------------------------------------------------------------------
int pgp_ispacket(BUFFER *b)
{
  return ((b->data[b->ptr] >> 6) == 2 || (b->data[b->ptr] >> 6) == 3);
}
//---------------------------------------------------------------------------
int pgp_dearmor(BUFFER *in, BUFFER *out)
{
  BUFFER *line, *temp;
  int err;// = 0;
  int tempbuf = 0;

  line = buf_new();
  temp = buf_new();

  if (in == out) {
    out = buf_new();
    tempbuf = 1;
  }
  do
    if (buf_getline(in, line) == -1) {
      err = -1;
      goto end;
    }
  while (!strleft((char*)line->data, begin_pgp)) ;

  while (buf_getheader(in, temp, line) == 0) ;	/* scan for empty line */

  /*err = */decode(in, out);
  err = buf_getline(in, line);
  if (line->length == 5 && line->data[0] == '=')	/* skip CRC */
    err = buf_getline(in, line);
  if (err == 0 && line->length && strleft((char*)line->data, end_pgp))
    err = 0;
  else
    err = -1;

end:
  buf_free(temp);
  buf_free(line);

  if (tempbuf) {
    buf_move(in, out);
    buf_free(out);
  }
  return (err);
}
//---------------------------------------------------------------------------












